# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QButtonGroup,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QRadioButton,
    QVBoxLayout,
)


class FilterDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Filter data")
        vbox = QVBoxLayout(self)

        # filter type
        filter_type_groupbox = QGroupBox("Filter type")
        filter_type_layout = QHBoxLayout()

        # radio buttons
        self.lowpass_button = QRadioButton("Lowpass")
        self.lowpass_button.setChecked(True)
        self.highpass_button = QRadioButton("Highpass")
        self.bandpass_button = QRadioButton("Bandpass")
        self.notch_button = QRadioButton("Notch")

        self.filter_group = QButtonGroup(self)
        self.filter_group.addButton(self.lowpass_button)
        self.filter_group.addButton(self.highpass_button)
        self.filter_group.addButton(self.bandpass_button)
        self.filter_group.addButton(self.notch_button)

        filter_type_layout.addWidget(self.lowpass_button)
        filter_type_layout.addWidget(self.highpass_button)
        filter_type_layout.addWidget(self.bandpass_button)
        filter_type_layout.addWidget(self.notch_button)

        self.filter_group.buttonToggled.connect(self._on_filter_type_changed)

        filter_type_groupbox.setLayout(filter_type_layout)
        vbox.addWidget(filter_type_groupbox)

        # filter settings
        filter_settings_groupbox = QGroupBox("Filter settings")
        self.grid = QGridLayout()
        self.lower_label = QLabel("Lower cutoff frequency (Hz):")
        self.lower_edit = QDoubleSpinBox()
        self.lower_edit.setMinimum(0)
        self.lower_edit.setDecimals(2)
        self.lower_edit.setValue(1)
        self.lower_edit.setAlignment(Qt.AlignRight)
        self.upper_label = QLabel("Upper cutoff frequency (Hz):")
        self.upper_edit = QDoubleSpinBox()
        self.upper_edit.setMinimum(0)
        self.upper_edit.setDecimals(2)
        self.upper_edit.setValue(30)
        self.upper_edit.setAlignment(Qt.AlignRight)
        self.notch_label = QLabel("Notch frequency (Hz):")
        self.notch_edit = QDoubleSpinBox()
        self.notch_edit.setMinimum(0)
        self.notch_edit.setDecimals(2)
        self.notch_edit.setValue(50)
        self.notch_edit.setAlignment(Qt.AlignRight)

        self.grid.addWidget(self.lower_label, 0, 0)
        self.grid.addWidget(self.lower_edit, 0, 1)
        self.grid.addWidget(self.upper_label, 1, 0)
        self.grid.addWidget(self.upper_edit, 1, 1)
        self.grid.addWidget(self.notch_label, 2, 0)
        self.grid.addWidget(self.notch_edit, 2, 1)

        filter_settings_groupbox.setLayout(self.grid)
        vbox.addWidget(filter_settings_groupbox)

        # buttons
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.ok_button = self.buttonbox.button(QDialogButtonBox.Ok)
        vbox.addWidget(self.buttonbox)
        self.ok_button.setEnabled(False)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.lower_edit.valueChanged.connect(self.validate_inputs)
        self.upper_edit.valueChanged.connect(self.validate_inputs)
        self.notch_edit.valueChanged.connect(self.validate_inputs)

        self._param_widgets_config = {
            "lower": (self.lower_label, self.lower_edit),
            "upper": (self.upper_label, self.upper_edit),
            "notch": (self.notch_label, self.notch_edit),
        }

        self._filter_type_config = {
            self.lowpass_button: {"name": "Lowpass", "widgets": ["upper"]},
            self.highpass_button: {"name": "Highpass", "widgets": ["lower"]},
            self.bandpass_button: {"name": "Bandpass", "widgets": ["lower", "upper"]},
            self.notch_button: {"name": "Notch", "widgets": ["notch"]},
        }

        self._on_filter_type_changed(self.filter_group.checkedButton(), True)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    def _on_filter_type_changed(self, button, checked):
        if not checked:
            return

        config = self._filter_type_config[button]
        self.selected_filter_type = config["name"]
        visible_widget_keys = config["widgets"]

        # hide all parameter widgets first
        for key in self._param_widgets_config:
            for widget in self._param_widgets_config[key]:
                widget.setVisible(False)

        # show only the relevant ones for the selected filter type
        for key in visible_widget_keys:
            for widget in self._param_widgets_config[key]:
                widget.setVisible(True)

        self.validate_inputs()

    def validate_inputs(self):
        is_valid = False
        if self.selected_filter_type == "Bandpass":
            is_valid = self.upper_edit.value() > self.lower_edit.value()
        elif self.selected_filter_type == "Lowpass":
            is_valid = self.upper_edit.value() > 0
        elif self.selected_filter_type == "Highpass":
            is_valid = self.lower_edit.value() > 0
        elif self.selected_filter_type == "Notch":
            is_valid = self.notch_edit.value() > 0

        self.ok_button.setEnabled(is_valid)

    @property
    def lower(self):
        return (
            float(self.lower_edit.value())
            if self.selected_filter_type in ["Bandpass", "Highpass"]
            else None
        )

    @property
    def upper(self):
        return (
            float(self.upper_edit.value())
            if self.selected_filter_type in ["Bandpass", "Lowpass"]
            else None
        )

    @property
    def notch(self):
        return (
            float(self.notch_edit.value())
            if self.selected_filter_type == "Notch"
            else None
        )



# © MNELAB developers
#
# License: BSD (3-clause)

import importlib.util
from importlib import import_module

required = ["mne", "PySide6", "edfio", "matplotlib", "numpy", "pyxdf", "scipy", "pybv"]
optional = ["mne-qt-browser", "picard", "sklearn"]


have = {}
for package in required + optional:
    module_name = package.replace("-", "_")
    spec = importlib.util.find_spec(module_name)
    if spec is None:
        have[package] = False
    else:
        mod = import_module(module_name)
        try:
            version = getattr(mod, "__version__", "unknown")
        except Exception:
            version = "unknown"
        have[package] = version



from PySide6.QtWidgets import QCheckBox, QDialog, QDialogButtonBox, QVBoxLayout


class BrainVisionDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("Import BrainVision data")

        vbox = QVBoxLayout(self)
        self._ignore_marker_types = QCheckBox(
            "Ignore marker types (only use description)"
        )
        self._ignore_marker_types.setChecked(False)
        vbox.addWidget(self._ignore_marker_types)
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @property
    def ignore_marker_types(self):
        return self._ignore_marker_types.isChecked()



# © MNELAB developers
#
# License: BSD (3-clause)

import re
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import numpy as np
from mne import channel_type
from mne.channels import DigMontage
from mne.defaults import _handle_default


def count_locations(info):
    locs = np.array([ch["loc"][:3] for ch in info["chs"]])
    valid_locs = np.any(~np.isclose(locs, 0) & np.isfinite(locs), axis=1)
    return valid_locs.sum()


def image_path(fname):
    """Return absolute path to image fname."""
    root = Path(__file__).parent.parent
    return str((root / "images" / Path(fname)).resolve())


def natural_sort(lst):
    """Sort a list in natural order."""

    def key(s):
        return [
            int(t) if t.isdigit() else t.lower() for t in re.split(r"(\d+)", str(s))
        ]

    return sorted(lst, key=key)


def calculate_channel_stats(raw):
    # extract channel info
    nchan = raw.info["nchan"]
    data = raw.get_data()
    cols = defaultdict(list)
    cols["name"] = raw.ch_names
    cols["type"] = [channel_type(raw.info, i) for i in range(nchan)]

    # vectorized calculations
    cols["min"], cols["Q1"], cols["median"], cols["Q3"], cols["max"] = np.percentile(
        data, [0, 25, 50, 75, 100], axis=1
    )
    cols["mean"] = np.mean(data, axis=1)

    # scaling and units
    scalings = _handle_default("scalings")
    units = _handle_default("units")
    cols["unit"] = []
    for i in range(nchan):
        unit = units.get(cols["type"][i])
        scaling = scalings.get(cols["type"][i], 1)
        cols["unit"].append(unit if scaling != 1 else "")
        if scaling != 1:
            for col in ["min", "Q1", "mean", "median", "Q3", "max"]:
                cols[col][i] *= scaling

    return cols, nchan


@dataclass
class Montage:
    montage: DigMontage
    name: str
    path: Path = None



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

from mnelab.utils import calculate_channel_stats


class SortableTableWidgetItem(QTableWidgetItem):
    def __init__(self, value):
        super().__init__(str(value))
        self.value = value

    def __lt__(self, other):
        try:
            return float(self.value) < float(other.value)
        except ValueError:
            return str(self.value) < str(other.value)


class ChannelStats(QDialog):
    def __init__(self, parent, raw):
        super().__init__(parent=parent)

        # window
        self.setWindowTitle("Channel Stats")
        self.setMinimumSize(400, 300)
        layout = QVBoxLayout(self)

        # create table
        self.table = QTableWidget(self)
        self.table.setSortingEnabled(True)
        layout.addWidget(self.table)

        # populate table
        self.populate_table(raw)
        self.table.resizeColumnsToContents()

        # add close button
        buttonbox = QDialogButtonBox(QDialogButtonBox.Close)
        layout.addWidget(buttonbox)
        buttonbox.rejected.connect(self.reject)

        # calculate table width
        table_width = self.table.verticalHeader().width()
        table_width += sum(
            self.table.columnWidth(i) for i in range(self.table.columnCount())
        )
        table_width += self.table.verticalScrollBar().width()
        self.resize(table_width - 30, 550)
        self.setLayout(layout)

    def populate_table(self, raw):
        cols, nchan = calculate_channel_stats(raw)
        self.table.setColumnCount(9)
        self.table.setRowCount(nchan)
        self.table.setHorizontalHeaderLabels(
            [
                "Channel",
                "Name",
                "Type",
                "Unit",
                "Min",
                "Q1",
                "Mean",
                "Median",
                "Q3",
                "Max",
            ]
        )

        for i in range(nchan):
            item = QTableWidgetItem(str(i))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 0, item)

            item = QTableWidgetItem(cols["name"][i])
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 1, item)

            item = QTableWidgetItem(cols["type"][i].upper())
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 2, item)

            item = QTableWidgetItem(cols["unit"][i])
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.table.setItem(i, 3, item)

            for j, key in enumerate(["min", "Q1", "median", "Q3", "max"], start=4):
                formatted_value = f"{cols[key][i]:.2f}"
                item = QTableWidgetItem(formatted_value)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                self.table.setItem(i, j, item)



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QLabel,
    QVBoxLayout,
)


class PSDDialog(QDialog):
    def __init__(self, parent, fmin, fmax, montage):
        super().__init__(parent)
        self.setWindowTitle("Power Spectral Density")

        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        fmin_label = QLabel("Lower frequency (Hz):")
        self.fmin_input = QDoubleSpinBox()
        self.fmin_input.setMinimum(0)
        self.fmin_input.setDecimals(1)
        self.fmin_input.setValue(fmin)
        self.fmin_input.setSingleStep(1)
        self.fmin_input.setMaximum(fmax)
        self.fmin_input.setAlignment(Qt.AlignRight)
        grid.addWidget(fmin_label, 0, 0)
        grid.addWidget(self.fmin_input, 0, 1)

        fmax_label = QLabel("Upper frequency (Hz):")
        self.fmax_input = QDoubleSpinBox()
        self.fmax_input.setMinimum(0)
        self.fmax_input.setDecimals(1)
        self.fmax_input.setValue(60)
        self.fmax_input.setSingleStep(1)
        self.fmax_input.setMaximum(fmax)
        self.fmax_input.setAlignment(Qt.AlignRight)
        grid.addWidget(fmax_label, 1, 0)
        grid.addWidget(self.fmax_input, 1, 1)

        bad_label = QLabel("Include bad channels:")
        self.bad_checkbox = QCheckBox()
        self.bad_checkbox.setChecked(True)
        grid.addWidget(bad_label, 2, 0)
        grid.addWidget(self.bad_checkbox, 2, 1)

        color_label = QLabel("Use spatial colors:")
        self.color_checkbox = QCheckBox()
        self.color_checkbox.setChecked(True)
        grid.addWidget(color_label, 3, 0)
        grid.addWidget(self.color_checkbox, 3, 1)
        if not montage:
            color_label.setEnabled(False)
            self.color_checkbox.setEnabled(False)
            self.color_checkbox.setChecked(False)

        vbox.addLayout(grid)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.fmin_input.valueChanged.connect(self.fmax_input.setMinimum)
        self.fmax_input.valueChanged.connect(self.fmin_input.setMaximum)

        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @property
    def fmin(self):
        """Get the lower frequency limit."""
        return self.fmin_input.value()

    @property
    def fmax(self):
        """Get the upper frequency limit."""
        return self.fmax_input.value()

    @property
    def exclude(self):
        """Check if bad channels should be excluded."""
        if self.bad_checkbox.isChecked():
            return ()
        else:
            return "bads"

    @property
    def spatial_colors(self):
        """Check if spatial colors should be used."""
        return self.color_checkbox.isChecked()



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import QEvent, Qt, Signal
from PySide6.QtGui import QColor, QIcon, QPainter
from PySide6.QtWidgets import (
    QAbstractItemView,
    QFrame,
    QHeaderView,
    QPushButton,
    QTableWidget,
)

ROW_HEIGHT = 10


class SidebarTableWidget(QTableWidget):
    rowsMoved = Signal(int, int)  # custom signal emitted when drag-and-dropping rows

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setDefaultDropAction(Qt.MoveAction)
        self.setDropIndicatorShown(False)
        self.setDragDropOverwriteMode(False)
        self.setFrameStyle(QFrame.NoFrame)
        self.setFocusPolicy(Qt.NoFocus)
        self.setEditTriggers(QAbstractItemView.DoubleClicked)
        self.setColumnCount(3)
        self.setShowGrid(False)
        self.drop_row = -1

        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.horizontalHeader().setStretchLastSection(False)
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.Fixed)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(2, QHeaderView.Fixed)
        self.setColumnWidth(2, 20)
        self.resizeColumnToContents(0)

        self.setMouseTracking(True)
        self.viewport().installEventFilter(self)

    def mousePressEvent(self, event):
        item = self.itemAt(event.pos())
        if item:
            super().mousePressEvent(event)
        else:
            event.ignore()

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        elif event.source() == self:
            event.accept()
            super().dragEnterEvent(event)
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        elif event.source() == self:
            drop_row = self.indexAt(event.pos()).row()
            if drop_row == -1:
                drop_row = self.rowCount()
            if drop_row != self.drop_row:
                self.drop_row = drop_row
            event.accept()
            super().dragMoveEvent(event)
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        self.drop_row = -1
        self.viewport().update()
        super().dragLeaveEvent(event)

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.drop_row >= 0:
            painter = QPainter(self.viewport())
            if self.drop_row < self.rowCount():
                y = self.visualRect(self.model().index(self.drop_row, 0)).top()
            else:
                y = self.visualRect(self.model().index(self.rowCount() - 1, 0)).bottom()
            painter.drawLine(0, y, self.viewport().width(), y)

    def dropEvent(self, event):
        if event.mimeData().hasUrls():
            self.parent.event(event)
        elif event.source() == self:
            drop_row = self.indexAt(event.pos()).row()
            if drop_row == -1:
                drop_row = self.rowCount()
            selected_rows = sorted(
                index.row() for index in self.selectionModel().selectedRows()
            )
            if selected_rows and drop_row > selected_rows[-1]:
                drop_row -= len(selected_rows)

            self.drop_row = -1
            self.viewport().update()
            self.rowsMoved.emit(selected_rows[0], drop_row)
            event.setDropAction(Qt.IgnoreAction)
            event.accept()
        else:
            event.ignore()

    def style_rows(self):
        self.resizeColumnToContents(0)
        for i in range(self.rowCount()):
            self.resizeRowToContents(i)
            self.setRowHeight(i, ROW_HEIGHT)
            self.item(i, 0).setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.item(i, 0).setForeground(QColor("gray"))
            self.item(i, 0).setFlags(self.item(i, 0).flags() & ~Qt.ItemIsEditable)
            self.item(i, 1).setTextAlignment(Qt.AlignLeft | Qt.AlignVCenter)
            self.item(i, 1).setFlags(self.item(i, 1).flags() | Qt.ItemIsEditable)

    def update_vertical_header(self):
        row_count = self.rowCount()
        self.setVerticalHeaderLabels([str(i) for i in range(row_count)])

    def eventFilter(self, source, event):
        if source == self.viewport() and event.type() == QEvent.MouseMove:
            index = self.indexAt(event.pos())
            if index.isValid():
                self.showCloseButton(index.row())
            else:
                self.showCloseButton(-1)
        elif source == self.viewport() and event.type() == QEvent.Leave:
            self.showCloseButton(-1)

        return False

    def showCloseButton(self, row_index):
        for i in range(self.rowCount()):
            if i == row_index:
                delete_button = QPushButton(self)
                delete_button.setIcon(QIcon.fromTheme("close-data"))
                delete_button.setStyleSheet(
                    "background: transparent; border: none; margin: auto;"
                )
                delete_button.clicked.connect(
                    lambda _, index=row_index: self.parent.model.remove_data(index)
                )
                self.setCellWidget(row_index, 2, delete_button)
            else:
                self.removeCellWidget(i, 2)



# © MNELAB developers
#
# License: BSD (3-clause)

import io
import sys

if getattr(sys, "frozen", False):
    if sys.stdout is None:
        sys.stdout = io.StringIO()
    if sys.stderr is None:
        sys.stderr = io.StringIO()

import multiprocessing as mp
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("mnelab")
except PackageNotFoundError:
    __version__ = "unknown"


def main():
    mp.freeze_support()
    mp.set_start_method("spawn", force=True)

    from pathlib import Path

    import matplotlib
    from PySide6.QtCore import QEvent, QLoggingCategory, Qt
    from PySide6.QtGui import QIcon
    from PySide6.QtWidgets import QApplication

    from mnelab.mainwindow import MainWindow
    from mnelab.model import Model

    QLoggingCategory.setFilterRules("*.debug=false\n*.warning=false")

    matplotlib.use("QtAgg")

    class MNELAB(QApplication):
        """MNELAB application."""

        def __init__(self, argv):
            super().__init__(argv)
            self.mainwindow = None

        def event(self, event):
            if event.type() == QEvent.FileOpen:
                self.mainwindow.open_data(event.file())
                return True
            return super().event(event)

    app = MNELAB(sys.argv)
    app.setApplicationName("mnelab")
    app.setApplicationDisplayName("MNELAB")
    app.setDesktopFileName("mnelab")
    app.setOrganizationName("mnelab")
    if sys.platform.startswith("darwin"):
        app.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus, True)
        app.setWindowIcon(QIcon(f"{Path(__file__).parent}/icons/mnelab-logo-macos.svg"))
    else:
        app.setWindowIcon(QIcon(f"{Path(__file__).parent}/icons/mnelab-logo.svg"))
    if sys.platform.startswith("win"):
        app.setStyle("fusion")
    model = Model()
    model.view = MainWindow(model)
    app.mainwindow = model.view
    if len(sys.argv) > 1:  # open files from command line arguments
        for f in sys.argv[1:]:
            model.view.open_data(f)
    model.view.show()
    sys.exit(app.exec())



# © MNELAB developers
#
# License: BSD (3-clause)
from pathlib import Path

from mne import get_config_path
from PySide6.QtCore import QPoint, QSettings, QSize, QStandardPaths, QUrl, Slot
from PySide6.QtGui import QDesktopServices, Qt
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QSpinBox,
)

SETTINGS_PATH = str(
    Path(
        QStandardPaths.writableLocation(
            QStandardPaths.StandardLocation.AppConfigLocation
        )
    )
    / "mnelab.ini"
)


_DEFAULTS = {
    "max_recent": 6,
    "max_channels": 20,
    "recent": [],
    "toolbar": True,
    "statusbar": True,
    "size": QSize(700, 500),
    "pos": QPoint(100, 100),
    "plot_backend": "Matplotlib",
}


def _get_value(key):
    return QSettings(SETTINGS_PATH, QSettings.Format.IniFormat).value(
        key, defaultValue=_DEFAULTS[key], type=type(_DEFAULTS[key])
    )


def read_settings(key=None):
    """
    Read application settings.

    Parameters
    ----------
    key : str, optional
        Setting key to read, by default `None`.

    Returns
    -------
    str | dict
        If `key` is given, return the corresponding value. If key is `None`, return all
        settings in a dictionary.
    """
    if key is not None:
        if key not in _DEFAULTS:
            raise KeyError(f"Invalid setting key: {key}")
        return _get_value(key)
    return {key: _get_value(key) for key in _DEFAULTS}


def write_settings(**kwargs):
    """Write application settings."""
    settings = QSettings(SETTINGS_PATH, QSettings.Format.IniFormat)
    for key, value in kwargs.items():
        if key not in _DEFAULTS:
            raise KeyError(f"Invalid setting key: {key}")
        settings.setValue(key, value)


def clear_settings():
    """Clear all settings."""
    QSettings(SETTINGS_PATH, QSettings.Format.IniFormat).clear()


class SettingsDialog(QDialog):
    def __init__(self, parent, backends):
        super().__init__(parent)
        self.setWindowTitle("Settings")

        grid = QGridLayout(self)

        backend = read_settings("plot_backend")
        if backend not in backends:
            backend = _DEFAULTS["plot_backend"]
        grid.addWidget(QLabel("Plot backend:"), 0, 0)
        self.plot_backend = QComboBox()
        self.plot_backend.addItems(backends)
        self.plot_backend.setCurrentIndex(backends.index(backend))
        grid.addWidget(self.plot_backend, 0, 1)

        grid.addWidget(QLabel("Recent files:"), 1, 0)
        self.max_recent = QSpinBox()
        self.max_recent.setRange(5, 25)
        self.max_recent.setValue(read_settings("max_recent"))
        self.max_recent.setAlignment(Qt.AlignRight)
        grid.addWidget(self.max_recent, 1, 1)

        grid.addWidget(QLabel("Displayed channels:"), 2, 0)
        self.max_channels = QSpinBox()
        self.max_channels.setRange(1, 256)
        self.max_channels.setValue(read_settings("max_channels"))
        self.max_channels.setAlignment(Qt.AlignRight)
        grid.addWidget(self.max_channels, 2, 1)

        mnelab_label = QLabel(
            f'<i>Settings are stored in <a href="{SETTINGS_PATH}">'
            f"{SETTINGS_PATH}</a>.</i>"
        )
        mnelab_label.linkActivated.connect(self.open_path)
        grid.addWidget(mnelab_label, 3, 0, 1, 2)

        mne_config_path = get_config_path()
        mne_label = QLabel(
            f'<i>MNE-Python settings are stored in <a href="{mne_config_path}">'
            f"{mne_config_path}</a>.</i>"
        )
        mne_label.linkActivated.connect(self.open_path)
        grid.addWidget(mne_label, 4, 0, 1, 2)

        hbox = QHBoxLayout()
        self.buttonbox = QDialogButtonBox(
            QDialogButtonBox.Apply | QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.reset_button = self.buttonbox.addButton(
            "Reset to defaults", QDialogButtonBox.ResetRole
        )
        hbox.addWidget(self.buttonbox)
        grid.addLayout(hbox, 5, 0, 1, 2)

        self.reset_button.clicked.connect(self.reset_settings)
        self.buttonbox.button(QDialogButtonBox.Apply).clicked.connect(
            self.apply_settings
        )
        self.buttonbox.accepted.connect(self.on_ok_clicked)
        self.buttonbox.rejected.connect(self.reject)

        grid.setSizeConstraint(QGridLayout.SetFixedSize)

    @Slot(str)
    def open_path(self, path):
        """Open a path in the default application."""
        QDesktopServices.openUrl(QUrl.fromLocalFile(path))

    @Slot()
    def apply_settings(self):
        write_settings(
            max_recent=int(self.max_recent.text()),
            max_channels=int(self.max_channels.text()),
            recent=self.parent().recent,
            plot_backend=self.plot_backend.currentText(),
        )
        self.parent().recent = self.parent().recent[: read_settings("max_recent")]

    @Slot()
    def on_ok_clicked(self):
        self.apply_settings()
        self.accept()

    @Slot()
    def reset_settings(self):
        self.max_recent.setValue(_DEFAULTS["max_recent"])
        self.max_channels.setValue(_DEFAULTS["max_channels"])
        self.plot_backend.setCurrentIndex(
            self.plot_backend.findText(_DEFAULTS["plot_backend"])
        )
        self.parent().resize(_DEFAULTS["size"])
        self.parent().move(_DEFAULTS["pos"])
        self.parent().recent = []
        clear_settings()



# © MNELAB developers
#
# License: BSD (3-clause)

from pathlib import Path

from numpy.rec import fromarrays
from scipy.io import savemat


def write_fif(fname, raw):
    raw.save(fname, overwrite=True)


def write_set(fname, raw):
    """Export raw to EEGLAB .set file."""
    data = raw.get_data() * 1e6  # convert to microvolts
    fs = raw.info["sfreq"]
    times = raw.times
    ch_names = raw.info["ch_names"]
    chanlocs = fromarrays([ch_names], names=["labels"])
    events = fromarrays(
        [
            raw.annotations.description,
            raw.annotations.onset * fs + 1,
            raw.annotations.duration * fs,
        ],
        names=["type", "latency", "duration"],
    )
    savemat(
        fname,
        {
            "EEG": {
                "data": data,
                "setname": fname,
                "nbchan": data.shape[0],
                "pnts": data.shape[1],
                "trials": 1,
                "srate": fs,
                "xmin": times[0],
                "xmax": times[-1],
                "chanlocs": chanlocs,
                "event": events,
                "icawinv": [],
                "icasphere": [],
                "icaweights": [],
            }
        },
        appendmat=False,
    )


def write_bdf_edf(fname, raw):
    """Export raw to EDF file."""
    raw.export(fname)


def write_bv(fname, raw):
    """Export data to BrainVision EEG/VHDR/VMRK file (requires pybv)."""
    raw.export(fname=Path(fname).with_suffix(".vhdr"))


# this dict contains each supported file extension as a key; the corresponding value is
# a list with three elements: (1) the writer function, (2) the full file format name,
# and (3) a (comma-separated) string indicating the supported objects
writers = {
    ".bdf": [write_bdf_edf, "Biosemi Data Format", "raw"],
    ".edf": [write_bdf_edf, "European Data Format", "raw"],
    ".eeg": [write_bv, "BrainVision", "raw"],
    ".fif": [write_fif, "Elekta Neuromag", "raw,epoch"],
    ".fif.gz": [write_fif, "Elekta Neuromag", "raw,epoch"],
    ".set": [write_set, "EEGLAB", "raw"],
}


def write_raw(fname, raw):
    maxsuffixes = max([ext.count(".") for ext in writers.keys()])
    suffixes = Path(fname).suffixes
    for i in range(-maxsuffixes, 0):
        ext = "".join(suffixes[i:])
        if ext in writers.keys():
            return writers[ext][0](fname, raw)
    else:
        raise ValueError(f"Unknown file type '{suffixes}'.")



# © MNELAB developers
#
# License: BSD (3-clause)

from mnelab.utils.dependencies import have
from mnelab.utils.syntax import PythonHighlighter
from mnelab.utils.utils import (
    Montage,
    calculate_channel_stats,
    count_locations,
    image_path,
    natural_sort,
)



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QColor, QPainter
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QHeaderView,
    QLabel,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

ROW_HEIGHT = 10


class DragDropTableWidget(QTableWidget):
    def __init__(self, parent=None, items=None):
        super().__init__(parent)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.DragDrop)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setDefaultDropAction(Qt.MoveAction)
        self.setDropIndicatorShown(False)
        self.setDragDropOverwriteMode(False)
        self.drop_row = -1

        self.setColumnCount(2)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setShowGrid(False)
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.horizontalHeader().setStretchLastSection(True)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        if items is not None:
            self.setRowCount(len(items))
            for i, (idx, name) in enumerate(items):
                self.setItem(i, 0, QTableWidgetItem(str(idx)))
                self.setItem(i, 1, QTableWidgetItem(name))
            self.style_rows()

    def style_rows(self):
        for i in range(self.rowCount()):
            self.resizeRowToContents(i)
            self.setRowHeight(i, ROW_HEIGHT)
            self.item(i, 0).setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.item(i, 0).setForeground(QColor("gray"))
            self.item(i, 0).setFlags(self.item(i, 0).flags() & ~Qt.ItemIsEditable)
            self.item(i, 1).setTextAlignment(Qt.AlignLeft | Qt.AlignVCenter)
            self.item(i, 1).setFlags(self.item(i, 1).flags() & ~Qt.ItemIsEditable)

    def dragMoveEvent(self, event):
        drop_row = self.indexAt(event.pos()).row()
        if drop_row == -1:
            drop_row = self.rowCount()
        if drop_row != self.drop_row:
            self.drop_row = drop_row
        event.accept()
        super().dragMoveEvent(event)

    def dragLeaveEvent(self, event):
        self.drop_row = -1
        self.style_rows()
        event.accept()

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.drop_row >= 0:
            painter = QPainter(self.viewport())
            if self.drop_row < self.rowCount():
                y = self.visualRect(self.model().index(self.drop_row, 0)).top()
            else:
                y = self.visualRect(self.model().index(self.rowCount() - 1, 0)).bottom()
            painter.drawLine(0, y, self.viewport().width(), y)

    def dropEvent(self, event):
        source_table = event.source()
        drop_row = self.indexAt(event.pos()).row()
        if drop_row == -1:
            drop_row = self.rowCount()

        selected_rows = sorted(
            index.row() for index in source_table.selectionModel().selectedRows()
        )

        if selected_rows:
            row_data = [
                [
                    source_table.item(row, col).text()
                    for col in range(source_table.columnCount())
                ]
                for row in selected_rows
            ]

            if source_table == self:
                for row in selected_rows:
                    if row < drop_row:
                        drop_row -= 1

            for row in reversed(selected_rows):
                source_table.removeRow(row)

            for i, data in enumerate(row_data):
                self.insertRow(drop_row + i)
                for col, value in enumerate(data):
                    self.setItem(drop_row + i, col, QTableWidgetItem(value))
            self.style_rows()
            event.accept()
        self.drop_row = -1


class AppendDialog(QDialog):
    def __init__(self, parent, compatibles, title="Append data"):
        super().__init__(parent)
        self.setWindowTitle(title)

        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        grid.addWidget(QLabel("Source"), 0, 0, Qt.AlignCenter)
        grid.addWidget(QLabel("Destination"), 0, 2, Qt.AlignCenter)

        self.source = DragDropTableWidget(self, items=compatibles)

        self.move_button = QPushButton("→")
        self.move_button.setEnabled(False)
        grid.addWidget(self.move_button, 1, 1, Qt.AlignHCenter)

        self.destination = DragDropTableWidget(self)

        grid.addWidget(self.source, 1, 0)
        grid.addWidget(self.destination, 1, 2)
        vbox.addLayout(grid)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        vbox.addWidget(self.buttonbox)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)
        self.destination.model().rowsInserted.connect(self.toggle_ok_button)
        self.destination.model().rowsRemoved.connect(self.toggle_ok_button)
        self.source.itemSelectionChanged.connect(self.toggle_move_source)
        self.destination.itemSelectionChanged.connect(self.toggle_move_destination)
        self.move_button.clicked.connect(self.move)
        self.toggle_ok_button()
        self.toggle_move_source()
        self.toggle_move_destination()

    @property
    def selected_idx(self):
        selected = []
        for it in range(self.destination.rowCount()):
            index_item = self.destination.item(it, 0)
            if index_item:
                selected.append(int(index_item.text()))
        return selected

    @Slot()
    def toggle_ok_button(self):
        if self.destination.rowCount() > 0:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)

    @Slot()
    def toggle_move_source(self):
        if self.source.selectedItems():
            self.move_button.setEnabled(True)
            self.move_button.setText("→")
            self.destination.clearSelection()
        elif not self.destination.selectedItems():
            self.move_button.setEnabled(False)

    @Slot()
    def toggle_move_destination(self):
        if self.destination.selectedItems():
            self.move_button.setEnabled(True)
            self.move_button.setText("←")
            self.source.clearSelection()
        elif not self.source.selectedItems():
            self.move_button.setEnabled(False)

    @Slot()
    def move(self):
        source_table = self.source if self.source.selectedRanges() else self.destination
        destination_table = (
            self.destination if self.source.selectedRanges() else self.source
        )

        rows = sorted(
            index.row() for index in source_table.selectionModel().selectedRows()
        )

        for row in rows:
            idx_item = source_table.item(row, 0).clone()
            name_item = source_table.item(row, 1).clone()
            row_count = destination_table.rowCount()
            destination_table.insertRow(row_count)
            destination_table.setItem(row_count, 0, idx_item)
            destination_table.setItem(row_count, 1, name_item)

        for row in reversed(rows):
            source_table.removeRow(row)

        source_table.style_rows()
        destination_table.style_rows()



# © MNELAB developers
#
# License: BSD (3-clause)
from pathlib import Path

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from mne.channels import make_standard_montage, read_custom_montage
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QHBoxLayout,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from mnelab.utils import Montage

SUPPORTED_FILES = [
    "*.loc",
    "*.locs",
    "*.eloc",
    "*.sfp",
    "*.csd",
    "*.elc",
    "*.txt",
    "*.elp",
    "*.bvef",
    "*.csv",
    "*.tsv",
    "*.xyz",
]


class MontageItem(QListWidgetItem):
    def __init__(self, montage, name, path=None):
        super().__init__(name)
        self.montage = montage
        self.name = name
        self.path = path


class MontageDialog(QDialog):
    def __init__(self, parent, montages):
        super().__init__(parent)
        self.setWindowTitle("Set montage")
        self.resize(760, 500)

        vbox = QVBoxLayout()
        self.montages = QListWidget()
        self.montages.setSelectionMode(QListWidget.SingleSelection)
        for name in montages:
            montage = make_standard_montage(name)
            item = MontageItem(montage, name)
            self.montages.addItem(item)
        vbox.addWidget(self.montages)

        self.montages.itemSelectionChanged.connect(self.view_montage)
        self.match_case = QCheckBox("Match case-sensitive", self)
        self.match_alias = QCheckBox("Match aliases", self)
        self.ignore_missing = QCheckBox("Ignore missing", self)
        self.ignore_missing.setChecked(True)
        vbox.addWidget(self.match_case)
        vbox.addWidget(self.match_alias)
        vbox.addWidget(self.ignore_missing)

        hbox = QHBoxLayout()
        hbox.addLayout(vbox, stretch=1)
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas_container = QWidget(self)
        self.canvas_container.setMinimumWidth(240)
        self.canvas_container.setMinimumHeight(220)
        self.canvas_layout = QVBoxLayout(self.canvas_container)
        self.canvas_layout.setContentsMargins(0, 0, 0, 0)
        self.canvas_layout.setSpacing(0)
        self.canvas_layout.addWidget(self.canvas)
        self.canvas_container.setLayout(self.canvas_layout)
        hbox.addWidget(self.canvas_container, stretch=2)

        button_layout = QHBoxLayout()
        self.open_file_button = QPushButton("Load custom montage...", self)
        self.open_file_button.clicked.connect(self.load_custom_montage)
        button_layout.addWidget(self.open_file_button)
        button_layout.addStretch()
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        button_layout.addWidget(self.buttonbox, alignment=Qt.AlignLeft)

        main_layout = QVBoxLayout(self)
        main_layout.addLayout(hbox)
        main_layout.addLayout(button_layout)
        self.setLayout(main_layout)

        if self.montages.count() > 0:
            self.montages.setCurrentRow(0)
            self.view_montage()

    def accept(self):
        item = self.montages.selectedItems()[0]
        self.montage = Montage(item.montage, item.name, item.path)
        super().accept()

    def view_montage(self):
        item = self.montages.currentItem()
        if item:
            montage = item.montage
            self.figure.clear()
            ax = self.figure.add_subplot()
            montage.plot(show_names=True, show=False, axes=ax)
            ax.set_aspect("equal")
            self.figure.tight_layout(pad=0.5)
            ax.margins(0)
            self.canvas.draw()

    def load_custom_montage(self):
        file_filter = f"Supported Files ({' '.join(SUPPORTED_FILES)});;All Files (*)"
        fpath, _ = QFileDialog.getOpenFileName(self, "Open File", "", file_filter)
        if fpath:
            try:
                montage = read_custom_montage(fpath)
            except Exception as e:
                QMessageBox.critical(self, "Unsupported montage file", str(e))
            else:
                item = MontageItem(montage, Path(fpath).name, Path(fpath))
                self.montages.addItem(item)
                self.montages.setCurrentRow(self.montages.count() - 1)
